// WinApp.h
/////////////////////////////////////////////////////////////////////////////

#ifndef _WIN_APP_H_
#define _WIN_APP_H_

#include "Window.h"
#include <vector>
#include <set>
#include <sstream>


namespace Manah {
namespace Windows {

/*
class CModule : public CSelfAssertable {
public:
	class ResourceIdentifier {
	public:
		ResourceIdentifier(int id) : m_id(id) {}
		ResourceIdentifier(const TCHAR* lpszName) : m_id(MAKEINTRESOURCE(lpszName)) {}
	private:
		int	m_id;
		friend class CModule;
	};

	// RXgN^
public:
	explicit CModule(HMODULE hModule) : m_hModule(hModule) {}
	virtual ~CModule() {}

	// \bh
public:
	HCURSOR	CreateCursor(int xHotSpot, int yHotSpot, int nWidth, int nHeight, const void* pvAndPlane, const void* pvXorPlane);
	HICON	CreateIcon(int nWidth, int nHeight, BYTE cPlanes, BYTE cBitsPixel, const BYTE* lpbAndBits, const BYTE* lpbXorBits);
	bool	EnumResourceLanguages(const TCHAR* lpszType, const TCHAR* lpszName, ENUMRESLAMGPROC lpEnumFunc, LONG_PTR lParam);
	bool	EnumResourceNames(const TCHAR* lpszType, ENUMRESNAMEPROC lpEnumFunc, LONG_PTR lParam);
	bool	EnumResourceTypes(ENUMRESTYPE lpEnumProc, LONG_PTR lParam);
	HICON	ExtractIcon(const TCHAR* lpszExeFileName, UINT iIcon);
	HRSRC	FindResource(const TCHAR* lpszName, const TCHAR* lpszType);
	HRSRC	FindResource(const TCHAR* lpszName, const TCHAR* lpszType, WORD nLanguage);
	HCURSOR	LoadCursor(ResourceIdentifier id);
	HICON	LoadIcon(ResourceIdentifier id);
	HANDLE	LoadImage(const TCHAR* lpszName, UINT nType, int cxDesired, int cyDesired, UINT nLoadOptions);
	HMENU	LoadMenu(ResourceIdentifier id);
	HGLOBAL	LoadResource(HRSRC hResInfo);
	DWORD	SizeofResource(HRSRC hResInfo);

protected:
	HMODULE	m_hModule;
};
*/

// CMessageArguments class definition
/////////////////////////////////////////////////////////////////////////////

class CMessageArguments {
	friend class CWinApp;

	// RXgN^
public:
	CMessageArguments() {}
private:
	CMessageArguments(const CMessageArguments& rhs);

	// Zq
public:
	template<class T>
	CMessageArguments& operator %(const T& rhs) {
		std::basic_ostringstream<TCHAR> ss;
		ss << rhs;
		m_args.push_back(ss.str());
		return *this;
	}
private:
	operator =(const CMessageArguments& rhs);

	// f[^o
private:
	std::list<std::tstring> m_args;
};

// shortened
#define __MARGS__	Manah::Windows::CMessageArguments()


// CWinApp class definition
/////////////////////////////////////////////////////////////////////////////

class CWinApp : public CSelfAssertable {
	// RXgN^
public:
	CWinApp(const Controls::CWindow* pWindow = 0);
	virtual ~CWinApp();

	// \bh
public:
	static void		GetCommandLineArguments(const TCHAR* pszCmdLine, std::vector<std::tstring>& args);
	void			GetCommandLineArguments(std::vector<std::tstring>& args) const;	// these method just split
																					// arguments by space
	Controls::CWindow&	GetMainWindow() const throw(std::logic_error);
	int				LoadString(UINT nID, TCHAR* lpszString, int cchMax) const;
	std::tstring	LoadString(UINT nID, const CMessageArguments& args = CMessageArguments()) const;
	std::tstring	LoadMessage(DWORD dwID, const CMessageArguments& args = CMessageArguments()) const;
	HCURSOR			LoadCursor(UINT nCursorID) const;
	HCURSOR			LoadCursor(const TCHAR* lpszCursorName) const;
	HCURSOR			LoadStandardCursor(const TCHAR* lpszCursorName) const;
	HICON			LoadIcon(UINT nIconID) const;
	HICON			LoadIcon(const TCHAR* lpszIconName) const;
	HICON			LoadStandardIcon(const TCHAR* lpszIconName) const;
	HBITMAP			LoadBitmap(UINT nID) const;
	HBITMAP			LoadBitmap(const TCHAR* lpszBitmapName) const;
	HMENU			LoadMenu(UINT nID) const;
	HMENU			LoadMenu(const TCHAR* lpszMenuName) const;
	void			LoadAccelerators(UINT nID);
	void			LoadAccelerators(const TCHAR* lpszTableName);
	UINT			GetProfileInt(const TCHAR* lpszSectionName, const TCHAR* lpszKeyName, UINT nDefault);
	bool			WriteProfileInt(const TCHAR* lpszSectionName, const TCHAR* lpszKeyName, UINT nValue);
	const TCHAR*	GetProfileString(const TCHAR* lpszSectionName,
						const TCHAR* lpszKeyName, const TCHAR* lpszDefault = 0);
	bool			WriteProfileString(const TCHAR* lpszSectionName,
						const TCHAR* lpszKeyName, const TCHAR* lpszString);
	bool			GetProfileStruct(const TCHAR* lpszSectionName,
						const TCHAR* lpszKeyName, void* lpStruct, UINT uSize);
	bool			WriteProfileStruct(const TCHAR* lpszSectionName,
						const TCHAR* lpszKeyName, void* lpStruct, UINT uSize);

	virtual bool	InitApplication();
	virtual bool	InitInstance();
	virtual int		Run();
protected:
	virtual LRESULT	DispatchEvent(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam) = 0;
	virtual bool	PreTranslateMessage(const MSG& msg);
	void			PopModelessDialog(HWND hDlg);
	void			PushModelessDialog(HWND hDlg);

	// f[^o
protected:
	Controls::CWindow*	m_pMainWindow;
	HINSTANCE			m_hInstance;				// application instance
//	const TCHAR*		m_pszCmdLine;				// command line
	TCHAR				m_szModulePath[MAX_PATH];	// executing application path
	TCHAR				m_szINIPath[MAX_PATH];		// INI file path
	TCHAR				m_szHelpPath[MAX_PATH];		// help file (*.hlp, *.chm) path
	HACCEL				m_hAcceleratorTable;
	std::set<HWND>		m_modelessDialogs;			// modeless dialogs to be called IsDialogMessage
};


// CWinApp class implementation
/////////////////////////////////////////////////////////////////////////////

inline CWinApp::CWinApp(const Manah::Windows::Controls::CWindow* pWindow /* = 0 */)
		: m_pMainWindow(const_cast<Manah::Windows::Controls::CWindow*>(pWindow)), m_hAcceleratorTable(0) {
	TCHAR*	psz = 0;

	m_hInstance = ::GetModuleHandle(0);
	::GetModuleFileName(0, m_szModulePath, MAX_PATH);
	psz = std::_tcsrchr(m_szModulePath, _T('.'));
	std::_tcscpy(m_szINIPath, m_szModulePath);
	std::_tcscpy(m_szINIPath + (psz - m_szModulePath + 1), _T("ini"));
	std::_tcscpy(m_szHelpPath, m_szModulePath);
	std::_tcscpy(m_szHelpPath + (psz - m_szModulePath + 1), _T("chm"));
}

inline CWinApp::~CWinApp() {
}

inline void CWinApp::GetCommandLineArguments(const TCHAR* pszCmdLine, std::vector<std::tstring>& args) {
	assert(pszCmdLine != 0);
	TCHAR*		pszEach = new TCHAR[std::_tcslen(pszCmdLine) + 1];
	TCHAR		ch;
	std::size_t	i = 0;
	std::size_t	cchEach = 0;
	bool		bInQuote = false;

	args.clear();
	while(true) {
		ch = *(pszCmdLine + i);
		if(ch == 0) {	// terminator
			if(cchEach != 0) {
				*(pszEach + cchEach) = 0;
				args.push_back(pszEach);
			}
			break;
		}
		if(!bInQuote) {
			if(ch == _T(' ')) {
				if(cchEach != 0) {
					*(pszEach + cchEach) = 0;
					args.push_back(pszEach);
					cchEach = 0;
				}
			} else if(ch == _T('\"'))
				bInQuote = true;
			else {
				*(pszEach + cchEach) = ch;
				++cchEach;
			}
		} else {
			if(ch == _T('\"'))
				bInQuote = false;
			else {
				*(pszEach + cchEach) = ch;
				++cchEach;
			}
		}
		++i;
	}
	delete[] pszEach;
}

inline void CWinApp::GetCommandLineArguments(std::vector<std::tstring>& args) const {
	AssertValid();

	const TCHAR*	pszCmdLine = ::GetCommandLine();
	int				i = 0;
	TCHAR			ch;
	TCHAR*			psz;

	while(true) {
		ch = *(pszCmdLine + i);
		if(ch == 0) {
			args.clear();
			return;
		} else if(ch == _T('\"')) {
			psz = std::_tcschr(pszCmdLine + i + 1, _T('\"'));
			if(psz == 0) {
				args.clear();
				return;
			}
			i = psz - pszCmdLine + 1;
		} else if(ch == _T('\'')) {
			psz = std::_tcschr(pszCmdLine + i + 1, _T('\''));
			if(psz == 0) {
				args.clear();
				return;
			}
			i = psz - pszCmdLine + 1;
		} else if(ch == _T(' ')) {
			++i;
			break;
		} else
			++i;
	}
	GetCommandLineArguments(pszCmdLine + i, args);
}

inline Manah::Windows::Controls::CWindow& CWinApp::GetMainWindow() const {
	AssertValid();
	if(m_pMainWindow == 0)
		throw std::logic_error("The application does not have window.");
	return *m_pMainWindow;
}

inline UINT CWinApp::GetProfileInt(const TCHAR* lpszSectionName, const TCHAR* lpszKeyName, UINT nDefault) {
	AssertValid();
	return ::GetPrivateProfileInt(lpszSectionName, lpszKeyName, nDefault, m_szINIPath);
}

inline const TCHAR* CWinApp::GetProfileString(const TCHAR* lpszSectionName,
		const TCHAR* lpszKeyName, const TCHAR* lpszDefault /* = 0 */) {
	AssertValid();
	static TCHAR	szReturned[16384];

	::GetPrivateProfileString(lpszSectionName, lpszKeyName,
		(lpszDefault != 0) ? lpszDefault : _T(""), szReturned, _countof(szReturned), m_szINIPath);
	return szReturned;
}

inline bool CWinApp::GetProfileStruct(
	   const TCHAR* lpszSectionName, const TCHAR* lpszKeyName, void* lpStruct, UINT uSize) {
	AssertValid();
	return toBoolean(::GetPrivateProfileStruct(lpszSectionName, lpszKeyName, lpStruct, uSize, m_szINIPath));
}

inline bool CWinApp::InitApplication() {
	return true;
}

inline bool CWinApp::InitInstance() {
	return true;
}

inline void CWinApp::LoadAccelerators(UINT nID) {
	AssertValid();
	m_hAcceleratorTable = ::LoadAccelerators(m_hInstance, MAKEINTRESOURCE(nID));
}

inline void CWinApp::LoadAccelerators(const TCHAR* lpszTableName) {
	AssertValid();
	m_hAcceleratorTable = ::LoadAccelerators(m_hInstance, lpszTableName);
}

inline HBITMAP CWinApp::LoadBitmap(UINT nID) const {
	AssertValid();
	return ::LoadBitmap(m_hInstance, MAKEINTRESOURCE(nID));
}

inline HBITMAP CWinApp::LoadBitmap(const TCHAR* lpszBitmapName) const {
	AssertValid();
	return ::LoadBitmap(m_hInstance, lpszBitmapName);
}

inline HCURSOR CWinApp::LoadCursor(UINT nCursorID) const {
	AssertValid();
	return ::LoadCursor(m_hInstance, MAKEINTRESOURCE(nCursorID));
}

inline HCURSOR CWinApp::LoadCursor(const TCHAR* lpszCursorName) const {
	AssertValid();
	return ::LoadCursor(m_hInstance, lpszCursorName);
}

inline HMENU CWinApp::LoadMenu(UINT nID) const {
	AssertValid();
	return ::LoadMenu(m_hInstance, MAKEINTRESOURCE(nID));
}

inline HMENU CWinApp::LoadMenu(const TCHAR* lpszMenuName) const {
	AssertValid();
	return ::LoadMenu(m_hInstance, lpszMenuName);
}

inline std::tstring CWinApp::LoadMessage(DWORD dwID, const CMessageArguments& args /* = CMessageArguments() */) const {
	AssertValid();

	void*	psz = 0;
	TCHAR**	pp = new TCHAR*[args.m_args.size()];
	std::list<std::tstring>::const_iterator	it;
	std::size_t				  				i;

	for(it = args.m_args.begin(), i = 0; it != args.m_args.end(); ++it, ++i)
		pp[i] = const_cast<TCHAR*>(it->c_str());
	::FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_HMODULE
		| FORMAT_MESSAGE_ARGUMENT_ARRAY, m_hInstance, dwID, LANG_USER_DEFAULT,
		reinterpret_cast<TCHAR*>(&psz), 0, reinterpret_cast<va_list*>(pp));
	delete[] pp;
	if(psz != 0) {
		TCHAR*		psz_ = static_cast<TCHAR*>(psz);
		std::size_t	cch = std::_tcslen(psz_);
		if(cch >= 2 && psz_[cch - 2] == _T('\r') && psz_[cch - 1] == _T('\n'))
			psz_[cch - 2] = 0;
		std::tstring str = psz_;
		::LocalFree(psz);
		return str;
	} else
		return _T("");
}

inline HCURSOR CWinApp::LoadStandardCursor(const TCHAR* lpszCursorName) const {
	AssertValid();
	return ::LoadCursor(0, lpszCursorName);
}

inline HICON CWinApp::LoadIcon(UINT nIconID) const {
	AssertValid();
	return ::LoadIcon(m_hInstance, MAKEINTRESOURCE(nIconID));
}

inline HICON CWinApp::LoadIcon(const TCHAR* lpszIconName) const {
	AssertValid();
	return ::LoadIcon(m_hInstance, lpszIconName);
}

inline HICON CWinApp::LoadStandardIcon(const TCHAR* lpszIconName) const {
	AssertValid();
	return ::LoadIcon(0, lpszIconName);
}

inline int CWinApp::LoadString(UINT nID, TCHAR* lpszString, int cchMax) const {
	AssertValid();
	return ::LoadString(m_hInstance, nID, lpszString, cchMax);
}

inline std::tstring CWinApp::LoadString(UINT nID, const CMessageArguments& args /* = CMessageArguments() */) const {
	AssertValid();

	TCHAR	szString[1024];
	::LoadString(m_hInstance, nID, szString, 1024);
	if(args.m_args.empty())
		return std::tstring(szString);

	void*	psz = 0;
	TCHAR**	pp = new TCHAR*[args.m_args.size()];
	std::list<std::tstring>::const_iterator	it;
	std::size_t							    i;

	for(it = args.m_args.begin(), i = 0; it != args.m_args.end(); ++it, ++i)
		pp[i] = const_cast<TCHAR*>(it->c_str());
	::FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_STRING
		| FORMAT_MESSAGE_ARGUMENT_ARRAY, szString, 0, 0,
		reinterpret_cast<TCHAR*>(&psz), 0, reinterpret_cast<va_list*>(pp));
	delete[] pp;

	std::tstring	strResult = static_cast<TCHAR*>(psz);
	::LocalFree(psz);
	return strResult;
}

inline bool CWinApp::PreTranslateMessage(const MSG& msg) {
	return false;
}

inline void CWinApp::PopModelessDialog(HWND hDlg) {
	AssertValid();
	m_modelessDialogs.erase(hDlg);
}

inline void CWinApp::PushModelessDialog(HWND hDlg) {
	AssertValid();
	assert(toBoolean(::IsWindow(hDlg)));
	m_modelessDialogs.insert(hDlg);
}

inline int CWinApp::Run() {
	MSG		msg;
	int		f;
	bool	bIsDlgMsg;

	while(true) {
		bIsDlgMsg = false;
		f = ::GetMessage(&msg, 0, 0, 0);
		assert(f != -1);
		if(f == 0)
			return msg.wParam;
		for(std::set<HWND>::iterator it = m_modelessDialogs.begin(); it != m_modelessDialogs.end(); ++it) {
			if(!toBoolean(::IsWindow(*it)))	// auto pop
				it = m_modelessDialogs.erase(it);
			else if(toBoolean(::IsDialogMessage(*it, &msg))) {
				bIsDlgMsg = true;
				break;
			}
		}
		if(!bIsDlgMsg) {
			if(m_hAcceleratorTable != 0) {
				if(!toBoolean(::TranslateAccelerator(msg.hwnd, m_hAcceleratorTable, &msg))
						&& !PreTranslateMessage(msg)) {
					::TranslateMessage(&msg);
					::DispatchMessage(&msg);
				}
			} else if(!PreTranslateMessage(msg)) {
				::TranslateMessage(&msg);
				::DispatchMessage(&msg);
			}
		}
	}

	return 0;
}

inline bool CWinApp::WriteProfileInt(const TCHAR* lpszSectionName, const TCHAR* lpszKeyName, UINT nValue) {
	AssertValid();

	TCHAR	szWrite[12];
	wsprintf(szWrite, _T("%d"), nValue);
	return toBoolean(::WritePrivateProfileString(lpszSectionName, lpszKeyName, szWrite, m_szINIPath));
}

inline bool CWinApp::WriteProfileString(const TCHAR* lpszSectionName, const TCHAR* lpszKeyName, const TCHAR* lpszString) {
	AssertValid();
	return toBoolean(::WritePrivateProfileString(lpszSectionName, lpszKeyName, lpszString, m_szINIPath));
}

inline bool CWinApp::WriteProfileStruct(
	   const TCHAR* lpszSectionName, const TCHAR* lpszKeyName, void* lpStruct, UINT uSize) {
	AssertValid();
	return toBoolean(::WritePrivateProfileStruct(lpszSectionName, lpszKeyName, lpStruct, uSize, m_szINIPath));
}

} /* namespace Windows */
} /* namespace Manah */

#endif /* _WIN_APP_H_ */

/* [EOF] */